import createSagaMiddleware__default, { effects } from 'redux-saga';
import * as createSagaMiddleware from 'redux-saga';
export { createSagaMiddleware as saga };
export { effects } from 'redux-saga';
import React from 'react';
import { unstable_batchedUpdates, render } from 'react-dom';
import { Provider } from 'react-redux';
export { connect, useDispatch, useSelector, useStore } from 'react-redux';
import { createStore, combineReducers, applyMiddleware, compose } from 'redux';
import _, { some, isString, replace, keys, get, isEmpty, has, mapValues, isArray, forEach, isFunction, isPlainObject, extend, cloneDeepWith, reduce, set, noop, isElement, pickBy, isUndefined, cloneDeep } from 'lodash';
import invariant from 'invariant';
import update from 'immutability-helper-x';
import { createSelectorCreator, defaultMemoize } from 'reselect';

// Object type
const MODULE = 'MODULE';
const PRESENTER = Symbol('PRESENTER'); // Module type

const DEFAULT_MODULE = Symbol('DEFAULT_MODULE');
const CONTAINER_MODULE = Symbol('CONTAINER_MODULE');
const PAGE_MODULE = Symbol('PAGE_MODULE');
const DOMAIN_MODULE = Symbol('DOMAIN_MODULE');
const ENTITY_MODULE = Symbol('ENTITY_MODULE');
const GLOBAL_MODULE = Symbol('GLOBAL_MODULE'); // Action type

const ACTION_KIND = {
  REDUCER: Symbol('REDUCER'),
  EFFECT: Symbol('EFFECT'),
  CREATOR: Symbol('CREATOR')
};
const CHANGE_LOADING_REDUCER = '@@loading-update';
const GET_LOADING_SELECTOR = '__getLoadings'; // For test
const KOP_GLOBAL_STORE_REF = Symbol('KOP_GLOBAL_STORE_REF');
const KOP_GLOBAL_SELECTOR_LOOP_CHECK = Symbol('KOP_GLOBAL_SELECTOR_LOOP_CHECK');

/**
 * @file redux中store变更后, 强制批量更新react
 */
function batchedUpdateMiddleware() {
  return next => action => unstable_batchedUpdates(() => next(action));
}

var defaultMiddleWares = [batchedUpdateMiddleware];

function isModule(value) {
  return value && value[MODULE] === MODULE;
}
function isPresenter(value) {
  return value && value[PRESENTER] === PRESENTER;
}
function toStorePath(path) {
  return replace(path, /\//g, '.');
} // get state by path
// parent module's data aside in .base

function getStateByNamespace(state, namespace, initialState) {
  const path = toStorePath(namespace);
  const initialStateKeys = keys(initialState);
  const findState = get(state, path);

  if (findState === undefined) {
    throw Error(`Please check if you forget to add module ${path} `);
  }

  if (isEmpty(initialState)) {
    if (findState['@@loading']) {
      return findState;
    }

    return get(state, `${path}.base`); // not in base
  }

  let isModuleState = true;
  initialStateKeys.forEach(key => {
    if (!has(findState, key)) {
      isModuleState = false;
    }
  });
  if (isModuleState) return findState;
  return get(state, `${path}.base`);
} // decorate put and call for tracing

function decorateSagaEffects(sagaEffects, referrer = '') {
  function put(action) {
    return sagaEffects.put({ ...action,
      referrer
    });
  }

  function call(fn, ...params) {
    if (fn.__actionKind) {
      if (fn.__actionKind === ACTION_KIND.EFFECT) {
        return sagaEffects.call(fn.__kopFunction, {
          payload: params.length > 1 ? params : params[0]
        });
      }

      return sagaEffects.put(fn(params.length > 1 ? params : params[0]));
    }

    return sagaEffects.call(fn, ...params);
  }

  return { ...sagaEffects,
    put,
    call
  };
}
function generateSagaEnhancer(events = {}, namespace, authCheck) {
  function syncPut(action) {
    if (authCheck) authCheck(action, namespace);

    if (action.__actionKind === ACTION_KIND.EFFECT) {
      const res = function* () {
        if (action.__changeLoading) {
          yield effects.put(action.__changeLoading({
            key: action.__type,
            value: true
          }));
        }

        const result = yield effects.call(action.__kopFunction, {
          payload: action.payload
        });

        if (action.__changeLoading) {
          yield effects.put(action.__changeLoading({
            key: action.__type,
            value: false
          }));
        }

        return result;
      };

      return res();
    }

    return effects.put(action);
  }

  return {
    syncPut,
    ...events
  };
}
function hasDuplicatedKeys(obj, ...others) {
  return some(obj, (val, key) => some(others, compareItem => {
    if (isString(compareItem)) {
      return compareItem === key;
    }

    return key in compareItem;
  }));
}

window[KOP_GLOBAL_SELECTOR_LOOP_CHECK] = window[KOP_GLOBAL_SELECTOR_LOOP_CHECK] || {};
let currentState = null; // check if selector has circular dependency

const selectorLoopChecker = {
  start: (namespace, key) => {
    if (!window[KOP_GLOBAL_SELECTOR_LOOP_CHECK]) {
      // eslint-disable-next-line @typescript-eslint/camelcase
      window[KOP_GLOBAL_SELECTOR_LOOP_CHECK] = {};
    }

    if (window[KOP_GLOBAL_SELECTOR_LOOP_CHECK][namespace + key] === 1) {
      throw Error(`module ${namespace}'s selector ${key} has circular dependency`);
    } else {
      window[KOP_GLOBAL_SELECTOR_LOOP_CHECK][namespace + key] = 1;
    }
  },
  end: (namespace, key) => {
    delete window[KOP_GLOBAL_SELECTOR_LOOP_CHECK][namespace + key];
  }
};

function _getStateValue(key, currState, namespace, initialState) {
  let stateValue;

  if (key === 'getGlobalState') {
    stateValue = currState;
    invariant(stateValue, 'Please check the kop store');
  } else {
    stateValue = getStateByNamespace(currState, namespace, initialState);
    invariant(stateValue, `Please check the kop-module ${namespace} is added, or you have return wrong state in last reducer`);
  }

  return stateValue;
}

function initSelectorHelper(store) {
  currentState = store.getState();
  store.subscribe(() => {
    currentState = store.getState();
  });
}
function createSelectors(namespace, selectors, presenter, initialState) {
  const globalizeSelector = (selector, key) => (...params) => {
    const stateValue = _getStateValue(key, currentState, namespace, initialState);

    selectorLoopChecker.start(namespace, key);

    if (params[0] === currentState) {
      params.shift();
    }

    const res = presenter.loaded ? selector(stateValue, presenter.selectors, ...params) : selector(stateValue, null, ...params);
    selectorLoopChecker.end(namespace, key);
    return res;
  };

  return mapValues(selectors, globalizeSelector);
}

function _transformSelectors(namespace, selectors, initialState) {
  const globalizeSelector = (selector, key) => (...params) => {
    if (isArray(params) && params.length > 1 && params[0] === currentState) {
      params.shift();
    }

    const stateValue = _getStateValue(key, currentState, namespace, initialState);

    const res = selector(stateValue, {
      payload: params.length === 1 ? params[0] : params
    });
    return res;
  };

  return mapValues(selectors, globalizeSelector);
}

const createPageSelectors = _transformSelectors;
const createGlobalSelectors = _transformSelectors;
function createDomainSelectors(selectors, entities) {
  const globalizeSelector = selector => (...params) => {
    if (isArray(params) && params.length > 1 && params[0] === currentState) {
      params.shift();
    }

    const res = selector({
      entities,
      payload: params.length === 1 ? params[0] : params
    });
    return res;
  };

  return mapValues(selectors, globalizeSelector);
}
function createContainerSelectors(namespace, selectors, presenter, initialState) {
  const globalizeSelector = (selector, key) => (...params) => {
    if (isArray(params) && params.length > 1 && params[0] === currentState) {
      params.shift();
    }

    const stateValue = _getStateValue(key, currentState, namespace, initialState);

    let res;

    if (presenter.loaded) {
      res = selector(stateValue, {
        pageSelectors: presenter.selectors,
        payload: params.length === 1 ? params[0] : params
      });
    } else {
      res = selector(stateValue, {
        payload: params.length === 1 ? params[0] : params
      });
    }

    return res;
  };

  return mapValues(selectors, globalizeSelector);
}

let _store;

function createApp(createOpt = {}) {
  let app;
  const {
    initialReducer,
    onError
  } = createOpt; // internal map for modules

  const _modules = {};

  let _router = React.createElement("div", null);

  const _middleWares = [...defaultMiddleWares];
  const sagaMiddleware = createSagaMiddleware__default();

  _middleWares.push(sagaMiddleware);

  function hasSubModule(module) {
    let flag = false;
    forEach(module, value => {
      if (isModule(value)) {
        flag = true;
      }
    });
    return flag;
  }

  function _addModule(m) {
    invariant(!_modules[m.namespace], `kop nodule ${m.namespace} exists`);

    if (!isEmpty(m.entities)) {
      forEach(m.entities, e => {
        _addModule(e);
      });
    }

    _modules[m.namespace] = m;
  } // create reducer
  // http://redux.js.org/docs/recipes/reducers/RefactoringReducersExample.html


  function createReducer(initialState = {}, handlers, defaultReducer) {
    return (state = initialState, action) => {
      if (Object.prototype.hasOwnProperty.call(handlers, action.type) && isFunction(handlers[action.type])) {
        const handler = handlers[action.type];
        return handler(state, action);
      }

      if (defaultReducer && isFunction(defaultReducer)) {
        return defaultReducer(state, action);
      }

      return state;
    };
  }

  function loopCombineReducer(tree, combineRootreducer = true, parentNode) {
    const childReducer = mapValues(tree, node => {
      if (!isModule(node)) {
        return loopCombineReducer(node);
      }

      if (hasSubModule(node)) {
        const subModuleMap = pickBy(node, value => isModule(value));
        return loopCombineReducer(subModuleMap, true, node);
      }

      return createReducer(node._initialState, node._reducers, node._defaultReducer);
    });
    let result;

    if (isEmpty(parentNode)) {
      result = { ...childReducer
      };
    } else if (parentNode === 'root') {
      invariant(!initialReducer || isPlainObject(initialReducer), 'initialReducer should be object');
      const noDuplicatedKeys = !hasDuplicatedKeys(initialReducer, childReducer, 'router');
      invariant(noDuplicatedKeys, 'initialReducer has reduplicate keys with other reducers');
      result = { ...initialReducer,
        ...childReducer
      };
    } else {
      result = {
        base: createReducer(parentNode._initialState, parentNode._reducers, parentNode._defaultReducer),
        ...childReducer
      };
    }

    if (parentNode === 'root' && !combineRootreducer) return result;
    return combineReducers(result);
  }

  function addModule(module) {
    if (isArray(module)) {
      module.forEach(m => {
        _addModule(m);
      });
    } else {
      _addModule(module);
    }
  }

  function initInjectModules(presenter) {
    forEach(presenter.injectModules, name => {
      invariant(_modules[name], `please check the kop-module ${name} is added`);
      extend(_modules[name].presenter, {
        loaded: true,
        selectors: presenter.selectors,
        actions: presenter.actions
      });
    });
  }

  function createRootReducer(combineRootreducer = true) {
    const moduleData = cloneDeepWith(_modules);
    const moduleTree = reduce(moduleData, (result, value, key) => {
      const module = get(result, toStorePath(key));

      if (isModule(value)) {
        if (module) {
          return set(result, `${toStorePath(key)}.base`, value);
        }

        return set(result, toStorePath(key), value);
      }

      return result;
    }, {});
    return loopCombineReducer(moduleTree, combineRootreducer, 'root');
  }

  function addPage(pageModule, opt = {}) {
    const {
      containers
    } = opt;

    if (containers && containers.length > 0) {
      addModule(containers);
    }

    if (pageModule.injectModules && pageModule.injectModules.length > 0) {
      initInjectModules(pageModule);
    }

    addModule(pageModule);
  }

  function addDomain(module) {
    addModule(module);
  }

  const addGlobal = _addModule;

  function removeModule(module) {
    const _remove = m => {
      invariant(_modules[m.namespace], `error: the kop-module - ${m.namespace} is not existed`); // hack redux-devtools's bug

      if (m && m.actions && !isPresenter(m) && m.type !== DOMAIN_MODULE && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__) {
        _store.dispatch(m.actions.__reset());
      }

      delete _modules[m.namespace];

      _store.dispatch({
        type: `${m.namespace}/@@KOP_CANCEL_EFFECTS`
      });
    };

    if (isArray(module)) {
      module.forEach(m => {
        _remove(m);
      });

      _store.replaceReducer(createRootReducer());
    } else if (module) {
      _remove(module);

      _store.replaceReducer(createRootReducer());
    }
  }

  function addRouter(r) {
    _router = r;
  }

  function addMiddleWare(middleWare) {
    const add = m => {
      _middleWares.push(m);
    };

    add(middleWare);
  }

  function getStore() {
    return _store;
  }

  function createAppStore() {
    // inject chrome redux devtools
    let composeEnhancers;

    if (process.env.NODE_ENV !== 'production' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__) {
      composeEnhancers = window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__;
    } else {
      composeEnhancers = compose;
    }

    const enhancer = composeEnhancers(applyMiddleware(..._middleWares));
    const rootReducer = createRootReducer();
    return createStore(rootReducer, enhancer);
  }

  function getArguments(...args) {
    let domSelector = null;
    let callback = noop;

    if (args.length === 1) {
      // app.start(false) means jump render phase and return early
      if (args[0] === false) {
        return {
          domSelector: '',
          callback: noop,
          shouldRender: false
        };
      }

      if (isString(args[0])) {
        domSelector = args[0];
      }

      if (isFunction(args[0])) {
        callback = args[0];
      }
    }

    if (args.length === 2) {
      domSelector = args[0];
      callback = args[1];
    }

    return {
      domSelector,
      callback,
      shouldRender: true
    };
  }

  function renderAppElement(domSelector, callback, shouldRender) {
    const $elem = React.createElement(Provider, {
      store: _store
    }, _router); // skip render when shouldRender is false

    if (shouldRender) {
      if (isString(domSelector)) {
        render($elem, document.querySelector(domSelector), () => {
          callback();
        });
      } else if (isElement(domSelector)) {
        render($elem, domSelector, () => {
          callback();
        });
      } else {
        callback();
      }
    }

    return $elem;
  }

  function _onError(e) {
    if (!onError) {
      console.error(e);
    } else {
      onError(e);
    }
  } // create root saga


  function createSaga(module) {
    return function* () {
      if (isArray(module.effects)) {
        for (let i = 0, k = module.effects.length; i < k; i++) {
          const task = yield effects.fork(function* () {
            try {
              if (module.effects) {
                yield module.effects[i]();
              }
            } catch (e) {
              _onError(e);
            }
          });
          yield effects.fork(function* () {
            yield effects.take(`${module.namespace}/@@KOP_CANCEL_EFFECTS`);
            yield effects.cancel(task);
          });
        }
      }

      if (isArray(module.watchers)) {
        for (let i = 0, k = module.watchers.length; i < k; i++) {
          const task = yield effects.fork(function* () {
            try {
              if (module.watchers) {
                yield module.watchers[i]();
              }
            } catch (e) {
              _onError(e);
            }
          });
          yield effects.fork(function* () {
            yield effects.take(`${module.namespace}/@@KOP_CANCEL_EFFECTS`);
            yield effects.cancel(task);
          });
        }
      }
    };
  }

  function start(...args) {
    const {
      domSelector,
      callback,
      shouldRender
    } = getArguments(...args);
    _store = createAppStore();
    initSelectorHelper(_store);
    app._store = _store;
    window[KOP_GLOBAL_STORE_REF] = _store;
    app._modules = _modules;
    forEach(app._modules, m => {
      if (m.type === ENTITY_MODULE) {
        m._store = _store; // 提供domainModel的遍历action
      }
    });
    app._middleWares = _middleWares;
    app._router = _router;
    forEach(_modules, module => {
      sagaMiddleware.run(createSaga(module));
    });
    return renderAppElement(domSelector, callback, shouldRender);
  } // 集成模式初始化，返回 saga，reducer 等等


  function _init() {
    const rootReducer = createRootReducer(false);
    return {
      rootReducer,
      sagaMiddleware
    };
  } // 集成模式启动，由外部 Redux 控制 App 渲染流程


  function _run(store) {
    initSelectorHelper(store);
    app._store = store;
    window[KOP_GLOBAL_STORE_REF] = store;
    app._modules = _modules;
    forEach(app._modules, m => {
      if (m.type === ENTITY_MODULE) {
        m._store = store;
      }
    });
    app._router = _router;
    forEach(_modules, module => {
      sagaMiddleware.run(createSaga(module));
    });
  }

  app = {
    addModule,
    addPage,
    addDomain,
    addGlobal,
    addRouter,
    addMiddleWare,
    start,
    getStore,
    removeModule,
    _init,
    _run
  };
  return app;
}

/*
 * create action
 */
function getActionTypeWithNamespace(namespace, type) {
  return `${namespace}/${type}`;
}
function createActionCreators({
  namespace,
  module,
  actionCreators,
  moduleType,
  checkAuth
}) {
  const actions = {};

  const createActionWarpper = type => {
    return createAction({
      namespace,
      actionType: type,
      actionKind: ACTION_KIND.CREATOR,
      moduleType
    });
  };

  let dispatch; // wrap dispatch for layered call auth check

  if (process.env.NODE_ENV === 'development') {
    dispatch = action => {
      checkAuth(action, namespace);
      return window[KOP_GLOBAL_STORE_REF].dispatch(action);
    };
  } else {
    dispatch = window[KOP_GLOBAL_STORE_REF].dispatch;
  }

  const moduleActionCreators = actionCreators({
    actions: module.actions,
    selectors: module.selectors,
    createAction: createActionWarpper,
    dispatch
  });
  forEach({ ...moduleActionCreators
  }, (creator, type) => {
    invariant(!module.actions[type], `Module ${namespace} action ${type} duplicated`); // wrap actionCreator to add meta data for the returned action

    if (process.env.NODE_ENV === 'development') {
      actions[type] = actionCreatorWithMetaTag(creator, namespace, moduleType, type);
    } else {
      actions[type] = creator; // directly pass in user defined actionCreator on module.actions
    }
  });
  return {
    actionCreators: actions
  };
}

const actionCreatorWithMetaTag = (creator, namespace, moduleType, type) => {
  return (...args) => {
    const action = creator(...args);
    action.__namespace = namespace;
    action.__moduleType = moduleType;

    action.toString = () => getActionTypeWithNamespace(namespace, type);

    return action;
  };
};

function createAction({
  namespace,
  actionType,
  actionKind,
  kopFunction,
  changeLoading,
  moduleType
}) {
  const fixedType = getActionTypeWithNamespace(namespace, actionType);

  const action = (payload, meta, referrer, error) => ({
    payload,
    meta,
    referrer,
    error,
    type: fixedType,
    __namespace: namespace,
    __actionKind: actionKind,
    __type: actionType,
    __kopFunction: kopFunction,
    __changeLoading: changeLoading,
    __moduleType: moduleType // meta tag for auth check

  });

  action.type = fixedType;

  action.toString = () => fixedType; // override toString to get namespace and type


  action.__actionKind = actionKind;
  action.__kopFunction = kopFunction;
  return action;
}

function transformEffects({
  moduleEffects,
  normalActions,
  namespace,
  moduleType,
  effectLoading
}) {
  const sagaFns = [];
  const effectActions = {}; // saga action

  forEach({ ...moduleEffects
  }, (value, type) => {
    invariant(!normalActions[type], `Module ${namespace} action ${type} duplicated`);
    const actionType = getActionTypeWithNamespace(namespace, type);
    let generator;
    let helper = effects.takeEvery;
    let args = []; // extra args for saga

    if (isArray(value)) {
      generator = value[0];
      const helperName = value[1];
      helper = effects[helperName];
      invariant(helper, `Module ${namespace} effect ${type} use invalid helper ${helperName}`);

      if (!isUndefined(value[2])) {
        args = value.slice(2);
      }
    } else {
      generator = value;
    }

    sagaFns.push(function* () {
      // fix redux-saga's bug
      // https://github.com/redux-saga/redux-saga/issues/1482
      yield helper(actionType, function* (...params) {
        try {
          if (effectLoading) {
            if (DOMAIN_MODULE !== moduleType) {
              yield effects.put(normalActions[CHANGE_LOADING_REDUCER]({
                key: type,
                value: true
              }));
            }
          }

          yield generator(...params);

          if (effectLoading) {
            if (DOMAIN_MODULE !== moduleType) {
              yield effects.put(normalActions[CHANGE_LOADING_REDUCER]({
                key: type,
                value: false
              }));
            }
          }
        } catch (e) {
          if (effectLoading) {
            if (DOMAIN_MODULE !== moduleType) {
              yield effects.put(normalActions[CHANGE_LOADING_REDUCER]({
                key: type,
                value: false
              }));
            }
          }

          console.error(e);
        }
      }, ...args);
    });
    effectActions[type] = createAction({
      namespace,
      moduleType,
      actionType: type,
      actionKind: ACTION_KIND.EFFECT,
      kopFunction: generator,
      changeLoading: normalActions[CHANGE_LOADING_REDUCER]
    });
  });
  return {
    sagaFns,
    effectActions
  };
}

function createEffectsAndActions(namespace, normalActions, selectors, events, effectsOpt) {
  const allActions = {};
  const effectActions = {};
  const sagaEffects = decorateSagaEffects(effects);
  const moduleEffects = effectsOpt(allActions, selectors, sagaEffects, events);
  const sagas = [];
  forEach({ ...moduleEffects
  }, (value, type) => {
    invariant(!normalActions[type], `Module ${namespace} action ${type} duplicated`);
    const actionType = getActionTypeWithNamespace(namespace, type);
    let generator;
    let helper = effects.takeEvery; // default helper is takeEvery

    let args = [];

    if (isArray(value)) {
      const helperName = value[1]; // effect function

      generator = value[0]; // saga function

      helper = effects[helperName];
      invariant(helper, `Module ${namespace} effect ${type} use invalid helper ${helperName}`);

      if (!isUndefined(value[2])) {
        args = value.slice(2);
      }
    } else {
      generator = value;
    }

    sagas.push(function* () {
      // fix redux-saga's bug
      // https://github.com/redux-saga/redux-saga/issues/1482
      yield helper(actionType, generator, ...args);
    });
    effectActions[type] = createAction({
      namespace,
      actionType: type,
      actionKind: ACTION_KIND.EFFECT,
      kopFunction: generator
    });
  });
  extend(allActions, normalActions, effectActions); // merge actions

  return {
    effectActions,
    allActions,
    effects: sagas
  };
} // auth check for effects

const getAuthCheck = {
  // RULE: cannot call other page's action
  [PAGE_MODULE]: (action, currentNamespace) => {
    const actionModuleType = action.__moduleType;
    const actionNamespace = action.__namespace;
    const notOwnPageModuleAction = PAGE_MODULE === actionModuleType && currentNamespace !== actionNamespace;

    if (notOwnPageModuleAction) {
      throw new Error('Page module can dispatch own actions, container module actions and domain module actions~');
    }

    return true;
  },
  // RULE: cannot call other container and page's action
  [CONTAINER_MODULE]: (action, currentNamespace) => {
    const isPageModuleAction = PAGE_MODULE === action.__moduleType;
    const notOwnContainerActions = CONTAINER_MODULE === action.__moduleType && currentNamespace !== action.__namespace;

    if (isPageModuleAction || notOwnContainerActions) {
      throw new Error('Container module can only dispatch own actions or domain module actions');
    }

    return true;
  },
  // RULE: cannot call container, page and other domain's action
  [DOMAIN_MODULE]: (action, currentNamespace) => {
    const isPageOrContainerModuleAction = PAGE_MODULE === action.__moduleType || CONTAINER_MODULE === action.__moduleType;
    const notOwnDomainActions = DOMAIN_MODULE === action.__moduleType && currentNamespace !== action.__namespace;

    if (isPageOrContainerModuleAction || notOwnDomainActions) {
      throw new Error('Domain module can only dispatch own actions');
    }
  },
  // RULE: can only dipatch its own actions
  [GLOBAL_MODULE]: (action, currentNamespace) => {
    const valid = GLOBAL_MODULE === action.__moduleType && currentNamespace === action.__namespace;

    if (!valid) {
      throw new Error('Global module can only dispatch its own actions');
    }

    return true;
  }
}; // internal method for get effects and actions

function _getEffectsAndActions(namespace, normalActions, selectors, events, effectsOpt, moduleType, effectLoading) {
  const allActions = {};
  const {
    enhancer,
    sagaEffects
  } = getWrappedEffectOptions(moduleType, namespace, events);
  const moduleEffects = effectsOpt({
    enhancer,
    selectors,
    actions: allActions,
    sagaEffects
  });
  const {
    sagaFns,
    effectActions
  } = transformEffects({
    moduleEffects,
    normalActions,
    namespace,
    moduleType,
    effectLoading
  });
  extend(allActions, normalActions, effectActions); // merge actions

  return {
    effectActions,
    allActions,
    effects: sagaFns
  };
}

const createPageEffects = _getEffectsAndActions;
const createContainerEffects = _getEffectsAndActions;
const createGlobalEffects = _getEffectsAndActions;
function createDomainEffects(namespace, selectors, servicesOpt, entities, moduleType, module) {
  const {
    enhancer,
    sagaEffects
  } = getWrappedEffectOptions(moduleType, namespace, {});
  const moduleEffects = servicesOpt({
    selectors,
    enhancer,
    entities,
    services: module.services,
    sagaEffects
  });

  moduleEffects.__reset = () => {
    forEach(entities, v => {
      v.__reset();
    });
  };

  const {
    sagaFns,
    effectActions
  } = transformEffects({
    moduleEffects,
    normalActions: {},
    namespace,
    moduleType,
    effectLoading: false
  });
  return {
    actions: effectActions,
    effects: sagaFns
  };
}

function parseActionsAndReducers(namespace, reducerMap, moduleType) {
  const actions = {};
  const reducers = {};
  forEach(reducerMap, (reducer, actionType) => {
    const actionTypeWithNamespace = getActionTypeWithNamespace(namespace, actionType);
    reducers[actionTypeWithNamespace] = reducer;
    actions[actionType] = createAction({
      namespace,
      actionType,
      moduleType,
      actionKind: ACTION_KIND.REDUCER
    });
  });
  return {
    actions,
    reducers
  };
}
function getModuleState(state) {
  return state;
}
function getGlobalState(state) {
  return state;
}
function initInitialState(module, pkg) {
  const {
    initialState = {}
  } = pkg;
  module._initialState = cloneDeep(initialState);
  module._initialState['@@loading'] = {};
}
const initModule = (pkg, namespace, type) => {
  invariant(namespace.indexOf(',') === -1, 'Module name can not contain " , "');
  return {
    namespace,
    type,
    [MODULE]: MODULE,
    parent: {},
    presenter: {
      loaded: false
    }
  };
}; // wrap saga effects with check auth

function wrapEffectsFunc(authCheck, namespace) {
  const wrapEffects = {};
  forEach(effects, (value, key) => {
    if (key === 'put') {
      wrapEffects[key] = (...args) => {
        const action = args && args[0];
        authCheck(action, namespace);
        return value(...args);
      };
    } else {
      wrapEffects[key] = value;
    }
  });
  return wrapEffects;
}
function initActions(module, pkg, namespace, moduleType, authCheck) {
  const {
    initialState = {},
    reducers,
    effectLoading = false
  } = pkg; // internal reducer for reset state

  const __reset = () => ({ ...initialState,
    '@@loading': {}
  });

  const changeLoading = (state, {
    payload
  }) => {
    if (state['@@loading']) {
      return update(state, {
        '@@loading': {
          [payload.key]: {
            $set: payload.value
          }
        }
      });
    }

    const loading = {
      [payload.key]: payload.value
    };
    return update(state, {
      '@@loading': {
        $set: loading
      }
    });
  };

  const reducersWithDefault = { ...reducers,
    __reset,
    [CHANGE_LOADING_REDUCER]: changeLoading
  };

  const getEventCompleteType = name => `${namespace}/_EVENT_${name}_`;

  const emit = ({
    name,
    payload
  }) => effects.put({
    payload,
    type: getEventCompleteType(name)
  });

  const take = name => effects.take(getEventCompleteType(name));

  const actionsAndReducers = parseActionsAndReducers(namespace, reducersWithDefault, moduleType);
  const normalActions = actionsAndReducers.actions;
  let actionCreators = {};

  if (pkg.actionCreators) {
    actionCreators = initActionCreators(module, pkg, namespace, moduleType, authCheck);
  } // extract actions from effects


  if (isFunction(pkg.effects)) {
    let data = {};

    switch (moduleType) {
      case PAGE_MODULE:
        data = createPageEffects(namespace, normalActions, module.selectors, {
          emit,
          take
        }, pkg.effects, moduleType, effectLoading);
        break;

      case CONTAINER_MODULE:
        data = createContainerEffects(namespace, normalActions, module.selectors, {
          emit,
          take
        }, pkg.effects, moduleType, effectLoading);
        break;

      case GLOBAL_MODULE:
        data = createGlobalEffects(namespace, normalActions, module.selectors, {
          emit,
          take
        }, pkg.effects, moduleType, effectLoading);
        break;

      default:
        data = createEffectsAndActions(namespace, normalActions, module.selectors, {
          emit,
          take
        }, pkg.effects);
    }

    return {
      actions: data.allActions,
      effects: data.effects,
      _reducers: actionsAndReducers.reducers,
      event: getEventCompleteType,
      actionCreators
    };
  }

  return {
    _reducers: actionsAndReducers.reducers,
    actions: normalActions,
    event: getEventCompleteType,
    actionCreators
  };
}
function initSelectors(module, pkg, namespace, moduleType) {
  const {
    selectors = {},
    initialState = {}
  } = pkg;

  selectors[GET_LOADING_SELECTOR] = state => state['@@loading'] || {};

  let moduleSelectors;

  switch (moduleType) {
    case PAGE_MODULE:
      moduleSelectors = createPageSelectors(namespace, { ...selectors,
        getModuleState,
        getGlobalState
      }, initialState);
      break;

    case CONTAINER_MODULE:
      moduleSelectors = createContainerSelectors(namespace, { ...selectors,
        getModuleState,
        getGlobalState
      }, module.presenter, initialState);
      break;

    case GLOBAL_MODULE:
      moduleSelectors = createGlobalSelectors(namespace, { ...selectors,
        getModuleState,
        getGlobalState
      }, initialState);
      break;

    default:
      moduleSelectors = createSelectors(namespace, { ...selectors,
        getModuleState,
        getGlobalState
      }, module.presenter, initialState);
  }

  module.selectors = moduleSelectors;
}
function initActionCreators(module, pkg, namespace, moduleType, checkAuth) {
  const {
    actionCreators
  } = createActionCreators({
    namespace,
    module,
    actionCreators: pkg.actionCreators,
    moduleType,
    checkAuth
  });
  return actionCreators;
}
function getWrappedEffectOptions(moduleType, namespace, events) {
  const authCheck = getAuthCheck[moduleType];
  let enhancer = generateSagaEnhancer(events, namespace);
  let effectsFunc = effects;

  if (process.env.NODE_ENV === 'development') {
    effectsFunc = wrapEffectsFunc(authCheck, namespace);
    enhancer = generateSagaEnhancer(events, namespace, authCheck);
  }

  return {
    sagaEffects: effectsFunc,
    enhancer
  };
}

function createPageWatchers(actions, selectors, watchersOpt, checkAuth, namespace) {
  let sagaEffects = effects;
  let enhancer = generateSagaEnhancer({}, namespace);

  if (process.env.NODE_ENV === 'development') {
    sagaEffects = wrapEffectsFunc(checkAuth, namespace);
    enhancer = generateSagaEnhancer({}, namespace, checkAuth);
  }

  const watchers = watchersOpt({
    actions,
    selectors,
    enhancer,
    sagaEffects
  });
  const watcherFns = [];

  function takeWatch(value, key) {
    if (isArray(value)) {
      watcherFns.push(function* () {
        yield effects[value[1]](key, value[0]);
      });
    }

    if (isFunction(value)) {
      watcherFns.push(function* () {
        yield effects.takeEvery(key, value);
      });
    }
  }

  forEach(watchers, (value, key) => {
    const keyArr = key.split(',');

    if (keyArr.length > 1) {
      keyArr.forEach(k => {
        takeWatch(value, k);
      });
    } else {
      takeWatch(value, key);
    }
  });
  return watcherFns;
}

function initInjectModules(module, pkg) {
  const {
    injectModules = []
  } = pkg;
  module.injectModules = injectModules.map(m => m.namespace);
}

function initPageWatchers(module, pkg) {
  const {
    namespace
  } = module;

  if (isFunction(pkg.watchers)) {
    module.watchers = createPageWatchers(module.actions, module.selectors, pkg.watchers, getAuthCheck[PAGE_MODULE], namespace);
  }
}

function createPageModule(namespace, pkg) {
  const module = initModule(pkg, namespace, PAGE_MODULE);
  initInitialState(module, pkg);
  initSelectors(module, pkg, namespace, PAGE_MODULE);
  initInjectModules(module, pkg);
  module.actions = {};
  const {
    actions,
    effects,
    _reducers,
    event,
    actionCreators
  } = initActions(module, pkg, namespace, PAGE_MODULE, getAuthCheck[PAGE_MODULE]);
  extend(module.actions, actionCreators, actions);

  if (effects) {
    module.effects = effects;
  }

  module._reducers = _reducers;
  module._defaultReducer = pkg.defaultReducer;
  module.event = event;
  initPageWatchers(module, pkg);
  return module;
}

function createContainerModule(namespace, pkg) {
  const module = initModule(pkg, namespace, CONTAINER_MODULE);
  initInitialState(module, pkg);
  initSelectors(module, pkg, namespace, CONTAINER_MODULE);
  module.actions = {};
  const {
    actions,
    effects,
    _reducers,
    event,
    actionCreators
  } = initActions(module, pkg, namespace, CONTAINER_MODULE, getAuthCheck[CONTAINER_MODULE]);
  extend(module.actions, actionCreators, actions);
  module._reducers = _reducers;
  module._defaultReducer = pkg.defaultReducer;
  module.event = event;

  if (effects) {
    module.effects = effects;
  }

  return module;
}

function createModule(namespace, pkg, type) {
  const module = initModule(pkg, namespace, type);
  initInitialState(module, pkg);
  initSelectors(module, pkg, namespace, DEFAULT_MODULE);
  const {
    actions,
    effects,
    _reducers,
    event
  } = initActions(module, pkg, namespace, DEFAULT_MODULE, noop);
  module.actions = actions;

  if (effects) {
    module.effects = effects;
  }

  module._reducers = _reducers;
  module.event = event;
  return module;
}

const arrize = input => {
  if (_.isArray(input)) {
    return input;
  }

  return [input];
}; // use _.isEqual instead of === for argument change determine


const createDeepEqualSelector = createSelectorCreator(defaultMemoize, // default memoize cache size is 1
_.isEqual); // 实体合并函数，默认是简单的浅合并

const mergeEntity = (target, source) => _.assign({}, target, source); // 主函数 createDomainModal


function createDomainModel(namespace, idKey) {
  const mapIds = (ids, byId) => _.map(ids, id => byId[id]);

  const mapIdsMemoized = _.memoize(mapIds, ids => ids.join());

  const getSelector = createDeepEqualSelector([(state, id) => state, (state, id) => id], (state, id) => {
    if (_.isArray(id)) {
      // return _.map(id, key => state.byId[key]);
      return mapIdsMemoized(id, state.byId);
    }

    return state.byId[id];
  });
  const selectSelector = createDeepEqualSelector([(state, predicate) => state, (state, predicate) => predicate], (state, predicate) => {
    const arr = _.map(state.ids, id => state.byId[id]);

    if (predicate) {
      return _.filter(arr, predicate);
    }

    return arr;
  });
  const module = {
    /**
     * Kop initialState
     */
    initialState: {
      byId: {},
      ids: []
    },

    /**
     * Kop selectors
     */
    selectors: {
      /**
       * 通过 id 直接拿数据，找不到则为 undefined
       * 例如：
       *  BaseModel.actions.get('xxx');
       * 批量:
       * BaseModel.actions.get(['xxx', 'xxxsss'])
       *
       * @param state
       * @param id
       *
       * @return {Object| Array | undefined}
       */
      get: (state, selectors, id) => // ignore selectors and use reselect
      getSelector(state, id),

      /**
       * 基于lodash的filter实现，你可以从model中选出所有符合条件的实体
       *
       * @param state
       * @param predicate 如果不传predicate，则返回全部数据
       * 例如：
       * BaseModel.actions.select({id: 'xxxx'});
       * 获取全部
       * BaseModel.actions.select();
       *
       * @return {Array|*}
       */
      select: (state, selectors, predicate) => // ignore selectors and use reselect
      selectSelector(state, predicate)
    },

    /**
     * Kop reducers
     */
    // tslint:disable-next-line:object-literal-sort-keys
    reducers: {
      /**
       * 特别说明：每个实体都必须是object，不支持array格式，例如：
       *  支持 { a: 1, b: 2 },
       *  不支持 [1, 2]
       *
       * 插入或者更新实体，支持批量插入或更新
       * 例如：
       *  BaseModel.actions.insert({ id: 'xxxx', a: 1, b: 2 });
       * 批量:
       * BaseModel.actions.insert([
       *   { id: 'xxxx', a: 1, b: 2 },
       *   { id: 'xxxxsss', a: 1, b: 3 },
       * ])
       */
      insert: (state, {
        payload: data
      }) => {
        mapIdsMemoized.cache.clear();
        const arr = arrize(data);

        const ids = _.map(arr, idKey);

        const byId = _.keyBy(arr, idKey);

        return {
          byId: _.chain(state.byId).clone().assignWith(byId, (objValue, srcValue) => {
            if (objValue === undefined) {
              return srcValue;
            }

            return mergeEntity(objValue, srcValue);
          }).value(),
          ids: _.union(state.ids, ids)
        };
      },

      /**
       * 根据id删除实体，支持批量
       * 例如：
       *  BaseModel.actions.delete('xxxx');
       * 批量
       *  BaseModel.actions.delete(['xxxx', 'xxxxsss']);
       */
      // tslint:disable-next-line:object-literal-sort-keys
      delete: (state, {
        payload: data
      }) => {
        mapIdsMemoized.cache.clear();
        const arr = arrize(data);
        return update(state, {
          byId: {
            $set: _.omit(state.byId, ...arr)
          },
          ids: {
            $set: _.without(state.ids, ...arr)
          }
        });
      },

      /**
       * 更新实体，支持批量，如果id不存在，则不会插入
       * 例如：
       * BaseModel.actions.update({ [idKey]: 'xxxx', b: 10 });
       * 批量
       * BaseModel.actions.update([
       *   { [idKey]: 'xxxx', b: 10 },
       *   { [idKey]: 'xxxxsss', a: 'lingyi' },
       * ])
       */
      update: (state, {
        payload: data
      }) => {
        mapIdsMemoized.cache.clear();
        const arr = arrize(data);

        const byId = _.keyBy(arr, idKey);

        const re = update.$set(state, 'byId', _.mapValues(state.byId, (entity, key) => {
          if (byId[key]) {
            return mergeEntity(entity, byId[key]);
          }

          return entity;
        }));
        return re;
      },

      /**
       * 清空数据库
       * 例如：
       * BaseModel.actions.clear(state);
       */
      clear: () => {
        mapIdsMemoized.cache.clear();
        return {
          ids: [],
          byId: {}
        };
      }
    }
  };
  const res = createModule(namespace, module, ENTITY_MODULE);
  const {
    actions,
    selectors
  } = res;
  const newActions = {}; // 转化一层，不用put了

  _.forEach(actions, (action, key) => {
    newActions[key] = (...params) => {
      if (returnModule._store) {
        returnModule._store.dispatch(action(...params));
      }
    };
  });

  const returnModule = { ...res,
    ...newActions,
    ...selectors
  };
  return returnModule;
}

/**
 * 创建container module
 */

function createDomainModule(namespace, pkg) {
  const module = initModule(pkg, namespace, DOMAIN_MODULE); // 初始化 entities 为 DomainModel

  initEntity(module, pkg, namespace); // 向 selectors 中注入实际生成的 entities 实例

  initDomainSelectors(module, pkg); // 注入运行时的实际参数

  initDomainActions(module, pkg, namespace, DOMAIN_MODULE);
  return module;
}
/*
 * 初始化 domain actions
 * todo：修改了入参
 */

function initDomainActions(module, pkg, namespace, moduleType) {
  const {
    services
  } = pkg;
  module._reducers = _.noop;
  module.services = {};

  if (_.isFunction(services)) {
    const {
      effects,
      actions
    } = createDomainEffects(namespace, module.selectors, services, module.entities, moduleType, module);
    module.effects = effects;

    _.extend(module.services, actions);
  } else {
    // module.services = services
    _.extend(module.services, services);
  }
}
/*
 * 初始化container selectors
 */

function initDomainSelectors(module, pkg) {
  const {
    selectors = {}
  } = pkg;
  module.selectors = createDomainSelectors({ ...selectors,
    getModuleState,
    getGlobalState
  }, module.entities);
}
/**
 * 注入领域模型的实体部分
 * @param module 模块
 * @param pkg 模块的配置
 * @param namespace 模块的命名空间
 */


function initEntity(module, pkg, namespace) {
  const entities = {};
  const {
    entities: entityOps
  } = pkg;

  _.forEach(entityOps, (e, key) => {
    entities[key] = createDomainModel(`${namespace}/@@entity-${key}`, e);
  });

  module.entities = entities;
}

function createGlobalModule(namespace, pkg) {
  const type = GLOBAL_MODULE;
  const module = initModule(pkg, namespace, type);
  initInitialState(module, pkg);
  initSelectors(module, pkg, namespace, type);
  module.actions = {};
  const {
    actions,
    effects,
    _reducers,
    event,
    actionCreators
  } = initActions(module, pkg, namespace, type, getAuthCheck[type]);
  extend(module.actions, actionCreators, actions);
  module._reducers = _reducers;
  module.event = event;

  if (effects) {
    module.effects = effects;
  }

  return module;
}

const firstUpperCase = ([first, ...rest]) => first.toUpperCase() + rest.join('');

function autoGetSet(options) {
  const {
    initialState,
    selectors,
    reducers
  } = options;
  const autoSelectors = {};
  const autoReducers = {};

  _.forEach(initialState, (_value, key) => {
    autoSelectors[`get${firstUpperCase(key)}`] = state => state[key];

    autoReducers[`set${firstUpperCase(key)}`] = (state, {
      payload
    }) => update.$set(state, key, payload);
  });

  options.selectors = { ...autoSelectors,
    ...selectors
  };
  options.reducers = { ...autoReducers,
    ...reducers
  };
  return options;
}

export default createApp;
export { autoGetSet, createContainerModule, createDomainModule, createGlobalModule, createPageModule };
